
import time
import threading


class Node(threading.Thread):

    def __init__(self, db, collect_replication=False, defer_collection=False):
        super(Node, self).__init__()

        self.db = db
        self.collect_repl_info = collect_replication

        self.num_sleepers = 0
        self.hostname = None
        self.connected_threads = 0
        self.max_connections = 0
        self.long_query_time = 0
        self.slave_status = {}
        self.row = {}

        if defer_collection == False:
            self.update()


    def update(self):
        self.num_sleepers = self.__get_num_sleepers()
        self.hostname = self.__get_hostname()
        self.connected_threads = self.__get_connected_threads()
        self.max_connections = self.__get_max_connections()
        self.long_query_time = self.__get_long_query_time()

        if self.collect_repl_info:
            self.slave_status = self.__get_slave_status()

        return self


    def __get_num_sleepers(self):
        sql = """
            SELECT count(id) AS num_sleepers 
            FROM information_schema.processlist 
            WHERE (command = 'Sleep' OR state = 'User sleep')
            """
        cur = self.db.query(sql)
        res = cur.fetchone()
        if res and 'num_sleepers' in res:
            return int(res['num_sleepers'])
        return 0


    def __get_connected_threads(self):
        sql = "SHOW GLOBAL STATUS LIKE 'Threads_connected'"
        cur = self.db.query(sql)
        res = cur.fetchone()
        if res and 'Value' in res:
            return int(res['Value'])
        return 0


    def __get_slave_status(self):
        sql = "SHOW SLAVE STATUS"
        cur = self.db.query(sql)
        return cur.fetchone()


    def __get_max_connections(self):
        if self.max_connections:
            return self.max_connections

        sql = "SHOW GLOBAL VARIABLES LIKE 'max_connections'"
        cur = self.db.query(sql)
        res = cur.fetchone()
        if res and 'Value' in res:
            return int(res['Value'])
        return 0


    def __get_hostname(self):
        if self.hostname:
            return self.hostname

        ## we're going to ask the remote mysql server
        sql = "SELECT @@hostname AS hostname"
        cur = self.db.query(sql)
        res = cur.fetchone()
        if res and 'hostname' in res:
            return res['hostname']

        return None


    def __get_long_query_time(self):
        if self.long_query_time:
            return self.long_query_time

        sql = "SHOW GLOBAL VARIABLES LIKE 'long_query_time'"
        cur = self.db.query(sql)
        res = cur.fetchone()
        if res and 'Value' in res:
            return int(round(float(res['Value'])))

        return 0


class ProcessNode(Node):

    def __init__(self, thread_lock, db, sql, thread_sleep_time=2):
        super(ProcessNode, self).__init__(db)
        self.thread_lock = thread_lock
        self.sql = sql
        self.thread_sleep_time = thread_sleep_time
        self.blocking = False
        self.process_list = {}
        self.num_processes = 0


    def update(self):
        super(ProcessNode, self).update()
        self.process_list = self.__get_process_list()

        return self


    def run(self):
        self.thread_lock.acquire(self.blocking)
        while True:
            self.update()
            time.sleep(self.thread_sleep_time)


    def __get_process_list(self):
        try:
            cur = self.db.query(self.sql)
            res = cur.fetchall()

            self.num_processes = cur.rowcount
            return res
        except AttributeError:
            return None





import os
import ConfigParser
import time
import smtplib
import subprocess
import json
from email.mime.text import MIMEText
from slackclient import SlackClient
from .exceptions import NotificationError

import pprint

pp = pprint.PrettyPrinter(indent=4)

# email hack sends subject (1) and body (2) to an external script in order to get around
# oddities when inside a container that would otherwise require us to setup/configure postfix/ssmtp etc...
SUPPORTED_NOTIFICATION_TYPES = ('email', 'emailhack', 'slack')

EXTERNAL_SENDMAIL = '/sendmail.sh'

'''
NOTE: There can be more than one slack channel. They must be separated by a comma.
----------------------------------------
`~/.slack_auth`
[slack]
username: jazzcibot
token: theSuperSecretToken
channels: #tech-jazzbot-noise,#second-optional-channelname
'''


class Notification(object):

    def __init__(self, proxies = None):
        self.proxies = proxies
        self.slack_config = None

        self.notification_type = os.getenv('NOTIFICATION_TYPE', None)

        if not self.notification_type or self.notification_type not in SUPPORTED_NOTIFICATION_TYPES:
            _m = '\n'.join(' - {t}'.format(t=t) for t in SUPPORTED_NOTIFICATION_TYPES)
            _m = '''NOTIFICATION_TYPE environment variable not set. Supported notification types:
{tstr}
'''.format(tstr=_m)

            raise NotificationError(_m)

        self.notification_type = self.notification_type.lower()

    def send(self, msg_short, msg_long, kwargs):
        try:
            if self.notification_type == 'email':
                self.send_email(msg_short, msg_long)

            if self.notification_type == 'slack':
                self.send_slack_msg(msg_long, msg_short, kwargs)

            if self.notification_type == 'emailhack':
                self.send_email_hack(msg_short, msg_long)

        except NotificationError as e:
            # let any raised NotificationError's bubbled up pass thru.
            # This means that if we blow a TypeError, or KeyError (etc) we'll have to catch that higher up
            print(e)
            pass


    def send_email_hack(self, subject, body):

        from_addr = os.getenv('NOTIFICATION_FROM', None)
        to_addr = os.getenv('NOTIFICATION_TO', None)
        relay = os.getenv('NOTIFICATION_HOST', None)
        port = os.getenv('NOTIFICATION_PORT', None)

        try:
            body = json.dumps(body, indent=2)
        except TypeError:
            pass

        cmd = [EXTERNAL_SENDMAIL, relay, port, from_addr, to_addr, '"{s}"'.format(s=subject), "'{b}'".format(b=body)]

        print("THE FOLLOWING IS THE NOTIFICATION COMMAND SENT TO THE SHELL SCRIPT")
        print(' '.join(cmd))

        proc = subprocess.Popen(
            ' '.join(cmd),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            executable='/bin/bash'
        )
        proc.wait()

        for line in proc.stdout.readlines():
            print(line)

        if proc.returncode != 0:
            return False

        return True


    def send_email(self, subject, body):
        from_addr = os.getenv('NOTIFICATION_FROM', None)
        to_addr = os.getenv('NOTIFICATION_TO', None)
        smtp_host = os.getenv('NOTIFICATION_HOST', 'localhost')

        if not from_addr or not to_addr:
            raise NotificationError('Either NOTIFICATION_{FROM|TO} is not set')

        # Give it our best shot to untangle things for the email
        if isinstance(body, dict):
            new_body = '\n'.join('{k}: {v}'.format(k=key, v=val) for key, val in body.items())
        else:
            new_body = body

        try:
            msg = MIMEText(new_body)

            msg['Subject'] = subject
            msg['From'] = from_addr
            msg['To'] = to_addr

            s = smtplib.SMTP(smtp_host)
            s.sendmail(from_addr, [to_addr], msg.as_string())
            s.quit()
        except Exception as e:
            # catch anything that may have happened here (for now), and turn that
            # into a NotificationError
            raise NotificationError(e)

    def __get_slack_client(self, auth_file):
        config = ConfigParser.ConfigParser()
        config.read(auth_file)

        self.slack_config = {
            'username': config.get('slack', 'username'),
            'token': config.get('slack', 'token'),
            'channels': config.get('slack', 'channels').split(',')
        }

        return SlackClient(self.slack_config.get('token'), proxies=self.proxies)

    def send_slack_msg(self, query_data, msg, kwargs={}):

        slack_auth_file = os.path.join(os.path.expanduser('~'), '.slack_auth')

        if not os.path.exists(slack_auth_file):
            raise NotificationError('Slack auth/config file not found! {f}'.format(f=slack_auth_file))

        slack_client = self.__get_slack_client()
        channels = self.slack_config.get('channels')
        fields = []

        for key, val in query_data.items():
            fields.append({
                'title': key,
                'value': val,
                'short': False if key == 'info' else True
            })

        attachments = [{
            'color':    '#E42217',
            'fields':   fields,
            'footer':   'mypsl query killah',
            'ts':       time.time()
        }]

        for channel in channels:
            resp = slack_client.api_call(
                'chat.postMessage',
                channel=channel,
                username=self.slack_config.get('username'),
                text=msg,
                icon_emoji=kwargs['icon'] if kwargs.get('icon') else None,
                attachments=attachments
            )

            if not resp.get('ok'):
                raise NotificationError('Notification.send failed for channel ({c}): {e}'.format(c=channel, e=resp.get('error')))




from distutils.spawn import find_executable
import pymysql
import os
import subprocess

import outputter as op
from .exceptions import MyDBError


def find_my_cnf():
    locations = (
        '/etc/my.cnf',
        '/etc/mysql/my.cnf',
        '/etc/mysql/conf.d/my.cnf',
        '/usr/etc/my.cnf',
    )

    for l in locations:
        if os.path.isfile(l):
            return l
    return None

def get_mysql_default(search_opt):
    my_print_defaults   = find_executable('my_print_defaults')
    my_cnf_file         = find_my_cnf()

    if not my_cnf_file:         return None
    if not my_print_defaults:   return None

    cmd     = [my_print_defaults, '--defaults-file', my_cnf_file, 'mysqld']
    proc    = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    proc.wait()

    if proc.returncode != 0:
        print(op.cv(proc.stderr.readlines(), op.Fore.RED + op.Style.BRIGHT))
        return None

    defaults = proc.stdout.readlines()

    if not defaults:
        return None

    for d in defaults:
        if d.startswith("--{0}".format(search_opt)):
            return d.split('=')[1].strip()
    return None


class mydb():
    conn            = None
    cursor          = None
    connect_args    = {}

    def __init__(self, margs, debug=False):
        self.debug = debug

        self.connect_args = {
            'db':           'information_schema',
            'cursorclass':  pymysql.cursors.DictCursor,
        }

        self.connect_args.update(margs)

        if self.connect_args['host'] == 'localhost':
            socket = get_mysql_default('socket')
            if socket:
                self.connect_args['unix_socket'] = socket
                del self.connect_args['host']
                del self.connect_args['port']

        pymysql.paramstyle = 'pyformat'


    def connect(self):
        try:
            self.conn = pymysql.connect(**self.connect_args)
        except pymysql.Error as e:
            raise MyDBError("Unable to connect. MySQL Said: {0}: {1}".format(e.args[0],e.args[1]))


    def query(self, sql, args=[]):
        try:
            if not self.cursor:
                self.cursor = self.conn.cursor()

            if args:
                self.cursor.execute(sql, args)
            else:
                self.cursor.execute(sql)

        except (AttributeError, pymysql.OperationalError):
            self.connect()

            if self.debug == True:
                print('pymysql exception, retrying...')

            self.query(sql, args)

        if self.cursor:
            return self.cursor

        return False


    def close(self):
        self.cursor_close()
        self.db_close()


    def cursor_close(self):
        if self.cursor:
            self.cursor.close()

    def db_close(self):
        if self.conn:
            self.conn.close()


import os
import yaml
from .exceptions import SaltClientError, SaltConfigError


def __load_from_config(config_dir, args, connect_dict):
    cfile = os.path.join(config_dir, args.connect_config)
    if os.path.isfile(cfile):
        with open(cfile, 'r') as f:
            connect_dict.update(yaml.load(f, Loader=yaml.FullLoader))
            return connect_dict

    return {}

def __load_from_salt(args, pillar_key='mysql:connection'):
    connect_args= {}
    try:
        import salt.client
    except ImportError:
        raise SaltConfigError('Unable to import salt.client')

    lc = salt.client.LocalClient()

    data = lc.cmd(args.salt_minion, 'network.ip_addrs')

    if not data:
        raise SaltClientError('Error: Unable to get the IP address via network.ip_addrs')

    try:
        connect_args['host'] = data[args.salt_minion][0]
    except (KeyError, IndexError, TypeError):
        raise SaltClientError("Error: Unable to get the IP address.")

    data = lc.cmd(args.salt_minion, 'pillar.get', [pillar_key])

    if not data:
        raise SaltClientError("Error: Unable to get {0} from pillar.".format(pillar_key))

    try:
        connect_args['user']   = data[args.salt_minion]['user']
        connect_args['passwd'] = data[args.salt_minion]['pass']
    except KeyError:
        raise SaltClientError("Error: Unable to get 'user' or 'pass' from pillar data.")

    return connect_args


'''
    Ok so what we're doing here is this....
    First, setup a dict of whatever came from `args`, which were parsed via the CLI - some have defaults.
    Next, update that with anything from the config file - if one was specified and found - config file will override
    defaults.

    At this point nothing is stopping the user from specifying --salt-minion as well, which will take precedence
'''
def prep_db_connection_data(config_dir, args):
    salt_connection = None

    db_connection = {
        'host': args.host,
        'passwd': args.passwd,
        'port': args.port,
        'user': args.user,
        'charset': args.charset
    }

    if args.connect_config:
        db_connection.update(__load_from_config(config_dir, args, db_connection))

    if args.salt_minion:
        salt_connection = __load_from_salt(args)

    if salt_connection:
        db_connection = salt_connection

    return db_connection




class Error(Exception):
    pass

class MyDBError(Error):
    pass

class SaltConfigError(Error):
    pass

class SaltClientError(Error):
    pass

class TooManyInstancesError(Error):
    pass

class NotificationError(Error):
    pass



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

'''
    mypsl :: MySQL process list watcher and query killer
    MIT License

    Copyright (c) 2018 Kyle Shenk

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
'''

from __future__ import print_function
import os
import threading
import sys
import argparse
import time

from libs.mysqldriver import mydb
from libs.processlist import ProcessList
from libs.processnode import ProcessNode
import libs.connections as connections
import libs.outputter as op

PROG_START = time.time()

try:
    import argcomplete
    HAS_ARGCOMPLETE = True
except ImportError:
    HAS_ARGCOMPLETE = False

INFO_TRIM_LENGTH        = 1000

'''
The following directory should contain files that are in yaml format
host: stuff.example.com
user: kshenk
passwd: things
port: 3306
'''
MYPSL_CONFIGS   = os.path.join(os.environ.get('HOME'), '.mypsl')

def _get_config_files(prefix, parsed_args, **kwargs):
    if not HAS_ARGCOMPLETE:
        return False
    if not os.path.isdir(MYPSL_CONFIGS):
        return False
    return next(os.walk(MYPSL_CONFIGS))[2]


def parse_args():
    parser = argparse.ArgumentParser(description=op.cv('(mypsl: {0}) :: MySQL Process list watcher & query killer.'.format(get_version()), op.Fore.CYAN + op.Style.BRIGHT),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    con_opt_group   = parser.add_argument_group(op.cv('Connection Options', op.Fore.YELLOW + op.Style.BRIGHT))
    config_group    = parser.add_argument_group(op.cv('Configuration Options', op.Fore.YELLOW + op.Style.BRIGHT))
    kill_group      = parser.add_argument_group(op.cv('Kill Options', op.Fore.RED + op.Style.BRIGHT))

    con_opt_group.add_argument('-H', '--host', dest='host', type=str, default='localhost',
        help='The host to get the process list from. If localhost, we will attempt to find and use the socket file first.')
    con_opt_group.add_argument('-p', '--port', dest='port', type=int, default=3306,
        help="The host's port. If the host is localhost, we will attempt to find and use the socket file first.")
    con_opt_group.add_argument('-u', '--user', dest='user', type=str, default='root',
        help='The user to connect to the host as.')
    con_opt_group.add_argument('-P', '--pass', dest='passwd', type=str, default='',
        help='The password for authentication.')
    #con_opt_group.add_argument('-S', '--socket', dest='socket', type=str,
    #    help='If connecting locally, optionally use this socket file instead of host/port.')
    con_opt_group.add_argument('-ch', '--charset', dest='charset', type=str, default='utf8',
        help='Charset to use with the database.')
    con_opt_group.add_argument('--config', dest='connect_config', type=str, default=None,
        help='Load connection configuration from a file in {0}. Just provide the filename. '.format(MYPSL_CONFIGS) + \
        'This will override any other connection information provided').completer = _get_config_files
    con_opt_group.add_argument('-sm', '--salt-minion', dest='salt_minion', type=str, default=None,
        help='Connect to mysql running on a salt minion. Do not use any other connection options with this. \
        mysql:connection:user and mysql:connection:pass must exist in pillar data.')

    ## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    config_group.add_argument('-l', '--loop', dest='loop_second_interval', type=int, default=0,
        help='Time in seconds between getting the process list.')
    config_group.add_argument('-lm', '--loop_max', dest='loop_max', type=int, default=None,
        help='If specified we will stop execution when the loop interval reaches this number')
    config_group.add_argument('-dft', '--default', dest='default', action='store_true',
        help='Run with defaults. Loop interval: 3 seconds, command like query or connect, order by time asc, id asc, truncate query to {0}.'.format(INFO_TRIM_LENGTH))
    config_group.add_argument('-c', '--command', dest='command', type=str,
        help='Lookup processes running as this command.')
    config_group.add_argument('-s', '--state', dest='state', type=str,
        help='Lookup processes running in this state.')
    config_group.add_argument('-t', '--time', dest='time', type=int,
        help='Lookup processes running longer than the specified time in seconds.')
    config_group.add_argument('-d', '--database', dest='database', type=str,
        help='Lookup processes running against this database.')
    config_group.add_argument('-q', '--query', dest='query', type=str,
        help='Lookup processes where the query starts with this specification.')
    config_group.add_argument('-i', '--id', dest='id_only', action='store_true',
        help='Only print back the ID of the processes.')
    config_group.add_argument('-isr', '--ignore_system_user', dest='ignore_system_user', action='store_true',
        help="Ignore the 'system user'")
    config_group.add_argument('--debug', dest='debug', action='store_true',
        help='Provide debug output.')
    config_group.add_argument('-o', '--order_by', dest='order_by', type=str,
        help='Order the results by a particular column: "user", "db asc", "db desc", "time desc"...etc')
    config_group.add_argument('-T', '--trim_info', dest='trim_info', action='store_true',
        help='Trim the info field (the query) to {0}'.format(INFO_TRIM_LENGTH))
    config_group.add_argument('-v', '--version', dest='version', action='store_true',
                              help='Show the installed program version and quit.')
    config_group.add_argument('--send-notification', dest='send_notification', action='store_true',
        help='Send a notification to slack anytime we kill a query. Slack Auth/Config file must exist @ ~/.slack_auth')

    ## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    kill_group.add_argument('--kill', dest='kill', action='store_true',
        help='Kill the queries that we find.')
    kill_group.add_argument('-kt', '--kill_threshold', dest='kill_threshold', default=100,
        help="The kill threshold. If a number is provided, we'll need to hit that many total connections before killing queries. You can \
        set this to 'off' as well, which kills queries no matter how many connections there are.")
    kill_group.add_argument('-ka', '--kill_all', dest='kill_all', action='store_true',
        help="If this flag is provided, we'll attempt to kill everything, not only select queries. {0}".format(op.cv("Use with caution!", op.Fore.RED + op.Style.BRIGHT)))
    kill_group.add_argument('-ky', '--kill_yes', dest='kill_yes', action='store_true',
        help="If this is provided we won't stop to ask if you are sure that you want to kill queries.")
    kill_group.add_argument('-kl', '--kill_log', dest='kill_log', default='/var/log/killed_queries.log',
        help="Where to log killed queries to, granting permissions to write to this file.")
    kill_group.add_argument('-klo', '--kill_log_only', dest='kill_log_only', action='store_true',
        help="Only log the queries that would be killed, do not actually kill them")

    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    return parser.parse_args()

def compile_sql(args):
    where = []
    order_by = []
    where_str = ''
    order_by_str = ''
    select_fields = ['id']

    if args.id_only and args.kill:
        print(op.cv("ERROR: Cannot specify id only (-i, --id) with kill!", op.Fore.RED + op.Style.BRIGHT))
        sys.exit(1)

    if args.kill and args.default:
        print(op.cv("ERROR: whoooaaaa no no no. You cannot kill using defaults!", op.Fore.RED + op.Style.BRIGHT))
        sys.exit(1)

    if args.kill:
        if not args.kill_yes:
            ans = raw_input(op.cv("Are you sure you want to kill queries? ", op.Style.BRIGHT))
            if ans.lower() not in ('y', 'yes'):
                print("Ok, then only use --kill when you are sure you want to kill stuff.")
                sys.exit(0)

    if not args.id_only:
        select_fields.extend(['user', 'host', 'db', 'command', 'time', 'state', 'info'])

    sql = "SELECT {0} FROM processlist".format(', '.join(select_fields))

    if args.default:
        where.append("(command = 'Query' OR command = 'Connect')")
        where.append("(command != 'Sleep')")
        args.loop_second_interval = 3
        args.ignore_system_user = True
        args.trim_info = True
        order_by = ['time ASC', 'id ASC']
    else:
        if args.command:
            where.append("command = '{0}'".format(args.command))
        if args.state:
            where.append("state = '{0}'".format(args.state))
        if args.time:
            where.append("time >= {0}".format(args.time))
        if args.database:
            where.append("db = '{0}'".format(args.database))
        if args.query:
            where.append("info LIKE '{0}%'".format(args.query))
        if args.order_by:
            order_by.append(args.order_by)

    if args.kill and not where:
        print(op.cv("ERROR: Cannot kill without specifying criteria!", op.Fore.RED + op.Style.BRIGHT))
        sys.exit(1)

    where.append("command != 'Binlog Dump'")
    where.append("(db != 'information_schema' OR db IS NULL)")  ## confuses me why I had to add OR db IS NULL

    if args.ignore_system_user == True:
        where.append("user != 'system user'")

    if where:
        where_str = 'WHERE {0}'.format(' AND '.join(where))

    if order_by:
        order_by_str = 'ORDER BY {0}'.format(', '.join(order_by))

    sql = ' '.join([sql, where_str, order_by_str])

    if args.debug:
        op.show_processing_time(PROG_START, time.time(), 'Program Preparation')
        print("SQL: {0}".format(op.cv(sql, op.Fore.CYAN)))

    return sql


def __shutdown(node_thread):
    try:
        node_thread.join()
    except RuntimeError:
        pass
    if node_thread.db:
        db = node_thread.db
        try:
            db.cursor_close()
            db.db_close()
        except Exception as e:
            print(op.cv(str(e), op.Fore.RED + op.Style.BRIGHT))

    print(op.cv('Quitting...', op.Fore.CYAN + op.Style.BRIGHT))
    print()
    sys.exit(0)


def establish_node(args, sql, threaded=False):
    db_auth = connections.prep_db_connection_data(MYPSL_CONFIGS, args)
    db = mydb(db_auth)
    db.connect()

    if args.debug:
        if db.conn:
            print(op.cv(
                ' --> db connection ({0}) established'.format(db_auth['host']),
                op.Fore.GREEN + op.Style.BRIGHT
            ))
        else:
            print(op.cv(
                ' --> db connection ({0}) failed'.format(db_auth['host']),
                op.Fore.RED + op.Style.BRIGHT
            ))

    pn = ProcessNode(threading.Lock(), db, sql)
    if threaded == True:
        pn.start()

    return pn


def __loop_control(counter, max):
    if max is None or isinstance(max, str):
        return True

    if max > 0:
        if counter < max:
            return True
        else:
            return False
    else:
        print(op.cv('The max loop interval should be greater than 0...'))
        sys.exit(1)


def display_process_lists(pl, loop_interval, loop_max):
    if loop_interval > 0:
        counter = 0
        while __loop_control(counter, loop_max):
            counter += 1
            if pl.process_processes(counter):
                counter = 0

            time.sleep(loop_interval)
            pl.update(time.time())
    else:
        pl.process_row()


def get_version():
    from libs._version import __version__
    return 'mypsl: {}'.format(__version__)


def main():
    args = parse_args()

    if args.version:
        print(get_version())
        sys.exit(0)

    sql = compile_sql(args)

    processNode = establish_node(args, sql).update()
    pl = ProcessList(processNode, vars(args))

    try:
        display_process_lists(pl, args.loop_second_interval, args.loop_max)
    except KeyboardInterrupt:
        __shutdown(processNode)

    try:
        processNode.join()
    except RuntimeError:
        pass

    sys.exit(0)


if __name__ == '__main__':
    main()




from collections import defaultdict
from .notification import Notification
from .exceptions import NotificationError
import os
import time
import outputter as op

READ_SEARCH     = ('show', 'select', 'desc')
WRITE_SEARCH    = ('insert', 'update', 'create', 'alter', 'replace', 'rename', 'delete')
LOCKED_SEARCH   = ('locked', 'waiting for table level lock', 'waiting for table metadata lock')

PROCESS_THRESHOLD_WARN  = 100
PROCESS_THRESHOLD_CRIT  = 200
SLEEPER_THRESHOLD_WARN  = 30
SLEEPER_THRESHOLD_CRIT  = 75
INFO_TRIM_LENGTH        = 1000

class ProcessList():

    def __init__(self, process_node, config):
        self.process_node = process_node
        self.start_time = time.time()
        self.config = config

    def update(self, start_time):
        self.start_time = start_time
        self.process_node = self.process_node.update()

        return self

    def print_header(self):
        ct = self.process_node.connected_threads
        mc = self.process_node.max_connections

        if (ct > (mc * .75)):
            ct_str = op.cv(ct, op.Fore.RED)
        elif (ct > (mc * .5)):
            ct_str = op.cv(ct, op.Fore.YELLOW)
        else:
            ct_str = op.cv(ct, op.Fore.CYAN)

        bar = "-" * 35
        header = "%s%s%s %s%s :: %s :: Threads (%s / %s) %s%s%s" % \
            (op.Fore.YELLOW, bar, op.Fore.GREEN, self.process_node.hostname, op.Fore.RESET, op.get_now_date(), ct_str, mc, op.Fore.YELLOW, bar, op.Fore.RESET)

        print(header)
        print("{0}".format(op.Style.BRIGHT) +
              op.OUT_FORMAT.format("ID", "USER", "HOST", "DB", "COMMAND", "TIME", "STATE", "INFO") +
              "{0}".format(op.Style.RESET_ALL))

    def process_row(self):
        num_reads = num_writes = num_locked = num_closing = num_opening = num_past_long_query = 0

        user_count          = defaultdict(int)
        state_count         = defaultdict(int)

        for row in self.process_node.process_list:
            if self.config.get('id_only'):
                print(row['id'])
                continue

            user_count[row['user']]     += 1
            state_count[row['state']]   += 1

            if row['info']:
                ## pull the first word of the query out - newline, tab or whitespace, and lowercase it.
                s_info = row['info'].split()[0].lower()
                if self.config.get('trim_info') and len(row['info']) > 1000:
                    row['info'] = "%s ..." % row['info'][:1000]
            else:
                row['info'] = '--'
                s_info = '--'

            if not row['state']:    row['state'] = '--'
            if not row['db']:       row['db'] = '--'

            ## the port number doesn't really tell us much.
            row['host'] = row['host'].split(':')[0]

            if s_info:
                if s_info in READ_SEARCH:   num_reads += 1
                if s_info in WRITE_SEARCH:  num_writes += 1

            if row['state']:
                if row['state'].lower() in LOCKED_SEARCH:           num_locked += 1
                if row['state'] == 'Copying to tmp table on disk':  num_writes += 1
                if row['state'].startswith('Opening table'):        num_opening += 1
                if row['state'].startswith('closing table'):        num_closing += 1

            if int(row['time']) > self.process_node.long_query_time: num_past_long_query += 1

            print \
                (op.OUT_FORMAT.format(row['id'], row['user'], row['host'], row['db'], row['command'], row['time'], row['state'], row['info'].encode('utf-8')))

        if self.config.get('id_only'):
            return

        return {
            'num_reads':            num_reads,
            'num_writes':           num_writes,
            'num_locked':           num_locked,
            'num_closing':          num_closing,
            'num_opening':          num_opening,
            'num_past_long_query':  num_past_long_query,
            'user_count':           user_count,
            'state_count':          state_count
        }


    def process_processes(self, counter=0):

        if not self.process_node.process_list:
            ## just sending a message to the terminal to let the user that the script is still working, and isn't stuck.
            if counter % 4 == 0:
                print(op.cv("{0} :: ({1}) :: Still looking...".format(op.get_now_date(), self.process_node.hostname), op.Style.BRIGHT))
            return False

        if self.config.get('kill'):
            kills = self.murdah()
            if kills:
                print(op.cv(op.get_now_date() + " :: " + self.process_node.hostname +
                        " :: Killed: " + str(kills) + " (SQL CRITERIA: {0})".format(self.process_node.sql), op.Fore.RED + op.Style.BRIGHT))
            return

        if not self.config.get('id_only'):
            self.print_header()

        _nums = self.process_row()

        if self.config.get('id_only'):
            ## then we're done here.
            return True

        num_reads           = _nums['num_reads']
        num_writes          = _nums['num_writes']
        num_locked          = _nums['num_locked']
        num_closing         = _nums['num_closing']
        num_opening         = _nums['num_opening']
        num_past_long_query = _nums['num_past_long_query']
        user_count          = _nums['user_count']
        state_count         = _nums['state_count']

        ## format total processes
        if self.process_node.num_processes >= PROCESS_THRESHOLD_CRIT:
            num_processes = op.cv(self.process_node.num_processes, op.Fore.RED)
        elif self.process_node.num_processes >= PROCESS_THRESHOLD_WARN:
            num_processes = op.cv(self.process_node.num_processes, op.Fore.YELLOW)
        else:
            num_processes = op.cv(self.process_node.num_processes, op.Fore.CYAN)

        ## format the number of queries past the long query time
        if num_past_long_query > 0:
            num_past_long_query = op.cv(num_past_long_query, op.Fore.RED)
        else:
            num_past_long_query = op.cv(num_past_long_query, op.Fore.CYAN)

        ## format the number of sleepers
        if self.process_node.num_sleepers >= SLEEPER_THRESHOLD_CRIT:
            num_sleepers = op.cv(self.process_node.num_sleepers, op.Fore.RED)
        elif self.process_node.num_sleepers >= SLEEPER_THRESHOLD_WARN:
            num_sleepers = op.cv(self.process_node.num_sleepers, op.Fore.YELLOW)
        else:
            num_sleepers = op.cv(self.process_node.num_sleepers, op.Fore.CYAN)

        print \
            ("\n\t({0}) PROCESSES: {1}, SLEEPERS: {2}, LOCKED: {3}, READS: {4}, WRITES: {5}, CLOSING: {6}, OPENING: {7}, PAST LQT: {8}"
              .format(op.cv(self.process_node.hostname, op.Fore.GREEN), num_processes, num_sleepers, op.cv(num_locked, op.Fore.CYAN),
                      op.cv(num_reads, op.Fore.CYAN), op.cv(num_writes, op.Fore.CYAN), op.cv(num_closing, op.Fore.CYAN),
                      op.cv(num_opening, op.Fore.CYAN), num_past_long_query))

        ## this is ok, but the next one sorts by occurrence
        # mystr = "{0}".format( ', '.join("%s: %s" % (k, "{0}".format(color_val(v, Fore.CYAN))) for (k, v) in user_count.iteritems()) )
        user_str = "{0}".format( ', '.join("%s: %s" % (k, "{0}".format(op.cv(user_count[k], op.Fore.CYAN))) \
                                           for k in sorted(user_count, key=user_count.get, reverse=True)) )
        state_str = "{0}".format(', '.join("%s: %s" % (k, "{0}".format(op.cv(state_count[k], op.Fore.CYAN))) \
                                           for k in sorted(state_count, key=state_count.get, reverse=True)))

        print("\t({0}) {1}".format(op.cv("Users", op.Fore.GREEN), user_str))
        print("\t({0}) {1}".format(op.cv("States", op.Fore.GREEN), state_str))

        print(op.show_processing_time(self.start_time, time.time()))

        print('')

        return True


    def record_kill(self, row):

        if self.config.get('send_notification'):
            proxies = {
                'http': 'http://proxy.svcs:3128',
                'https': 'http://proxy.svcs:3128'
            }
            msg = 'Query Killed on {hostname}'.format(hostname=self.process_node.hostname)

            # No need to bail if we just can't send the slack notification
            try:
                # Notification configuration is handled in the service definition
                notifier = Notification(proxies)
                # When sending an email notification, currently the 3rd argument (icon) doesn't have any affect
                notifier.send(msg, row, {'icon': ':jason:'})
            except Exception as e:
                # Anything that may have happened in there doesn't have to stop the show
                print(e)

        if os.path.exists(self.config.get('kill_log')):
            if not os.access(self.config.get('kill_log'), os.W_OK):
                return
        try:
            with open(self.config['kill_log'], 'a') as f:
                kill_string = "{0} :: {1} :: {2}\n".format(op.get_now_date(), self.process_node.hostname,
                                                           ', '.join(
                                                               "%s: %s" % (k, v) for (k, v) in row.items()))
                f.write(kill_string)

        except IOError:
            print(op.cv("Unable to write to: {0}".format(self.config['kill_log']), op.Fore.RED + op.Style.BRIGHT))
            return


    def murdah(self):
        ## ok. is it an integer and are the connected threads greater than the kill threshold ?
        try:
            kill_threshold = int(self.config.get('kill_threshold'))
            if self.process_node.connected_threads < kill_threshold:
                print("Connected threads: {0}, Kill threshold: {1}. Not killing at this time".format(
                    self.process_node.connected_threads, kill_threshold)
                )
                return
        except ValueError:
            if self.config.get('kill_threshold').lower() != 'off':
                ## if we haven't set this to off, then no killing
                print("kill threshold was set but doesn't = off. Not killing at this time.")
                return

        sql = "KILL %s"
        killed = 0
        for row in self.process_node.process_list:
            if not self.config.get('kill_all'):
                if not row['info'].lower().startswith('select'):
                    continue

            if self.config.get('kill_log_only'):
                self.record_kill(row)
            elif self.process_node.db.query(sql, (row['id'],)):
                self.record_kill(row)
                killed += 1

        return killed



from colorama import init, Fore, Style
import datetime

init()

OUT_FORMAT      = "{0:<12}{1:16}{2:20}{3:22}{4:25}{5:<8}{6:28}{7:25}"

def cv(val, color):
    return "%s%s%s" % (color, val, Style.RESET_ALL)

def get_now_date():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

def show_processing_time(start, end, text='Processing time'):
    elapsed = round(end - start, 3)

    if elapsed > 5:
        elapsed_str = cv(elapsed, Fore.RED)
    elif elapsed > .5:
        elapsed_str = cv(elapsed, Fore.YELLOW)
    else:
        elapsed_str = cv(elapsed, Fore.CYAN)

    return "\t({0}): {1}".format(cv(text, Fore.GREEN), elapsed_str)
